"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createVisitor = void 0;
const core_1 = require("@babel/core");
const names_1 = require("./names");
const findAndMarkRelevantFunctions_1 = require("./utils/findAndMarkRelevantFunctions");
const markUnsupported_1 = require("./utils/markUnsupported");
const findRootOfCallExpression_1 = require("./utils/findRootOfCallExpression");
const findRootDeclaration_1 = require("./utils/findRootDeclaration");
const transformGraphql_1 = require("./utils/transformGraphql");
const crypto_1 = require("crypto");
function verifyLoadQueryCall(path, arg, isDev) {
    if (!core_1.types.isIdentifier(arg)) {
        const msg = isDev
            ? `[GraphQL Compiler Plugin] ${name} must be passed an identifier as the first argument.`
            : `Invalid ${name} call (first argument was not an identifier)`;
        (0, markUnsupported_1.markWithError)(path, msg);
        return false;
    }
    return true;
}
function transformLoadQuery({ supported, unsupported }, nullify, isDev) {
    for (const loadQueryExpr of supported) {
        if (nullify) {
            loadQueryExpr.replaceWith(core_1.types.callExpression(loadQueryExpr.node.callee, [core_1.types.nullLiteral()]));
        }
        else {
            const arg = loadQueryExpr.node.arguments[0];
            if (verifyLoadQueryCall(loadQueryExpr, arg, isDev)) {
                const memberExpression = core_1.types.memberExpression(arg, core_1.types.identifier(names_1.names.literal));
                const callExpression = core_1.types.callExpression(memberExpression, []);
                loadQueryExpr.replaceWith(core_1.types.callExpression(loadQueryExpr.node.callee, [
                    callExpression,
                    ...loadQueryExpr.node.arguments.slice(1),
                ]));
            }
        }
    }
}
function verifyHook(path, arg, isDev) {
    if (!core_1.types.isTemplateLiteral(arg)) {
        const msg = isDev
            ? `[GraphQL Compiler Plugin] ${name} must be passed a template literal as the first argument.`
            : `Invalid ${name} call (first argument was not template literal)`;
        (0, markUnsupported_1.markWithError)(path, msg);
        return false;
    }
    return true;
}
function transformHooks(programPath, programHash, { supported, unsupported }, name, isFragment, hoist, rootNodesFound, isDev) {
    for (const hookExpr of supported) {
        const arg = hookExpr.node.arguments[0];
        if (verifyHook(hookExpr, arg, isDev)) {
            if (!hoist) {
                hookExpr.replaceWith(core_1.types.callExpression(hookExpr.node.callee, [core_1.types.nullLiteral()]));
            }
            else {
                const { parent, parentName } = (0, findRootDeclaration_1.findRootDeclaration)(hookExpr);
                const declarationRoot = (0, findRootOfCallExpression_1.findRootOfCallExpression)(hookExpr, name, isDev, rootNodesFound);
                if (declarationRoot == null) {
                    (0, markUnsupported_1.markWithError)(hookExpr, "Unable to find a root function declaration");
                    continue;
                }
                const memberExpression = core_1.types.memberExpression(core_1.types.identifier(parentName), core_1.types.identifier(names_1.names.literal));
                const memberInternalExpression = core_1.types.memberExpression(core_1.types.identifier(parentName), core_1.types.identifier(names_1.names.literInternalName));
                const callExpression = core_1.types.callExpression(memberExpression, []);
                hookExpr.replaceWith(core_1.types.callExpression(hookExpr.node.callee, [callExpression]));
                let nameExpression = core_1.types.stringLiteral("query");
                if (isFragment) {
                    const memberNameExpression = core_1.types.memberExpression(core_1.types.identifier(parentName), core_1.types.identifier(names_1.names.literName));
                    const fnHash = (0, crypto_1.createHash)("md5")
                        .update(parent.getSource())
                        .digest("hex")
                        .slice(0, 6);
                    const hoistedNameExpression = core_1.types.arrowFunctionExpression([], core_1.types.stringLiteral(parentName + "__" + programHash + "_" + fnHash));
                    const assignNameExpression = core_1.types.expressionStatement(core_1.types.assignmentExpression("=", memberNameExpression, hoistedNameExpression));
                    declarationRoot.insertAfter(assignNameExpression);
                    nameExpression = core_1.types.callExpression(memberNameExpression, []);
                }
                const { ast, referenced } = (0, transformGraphql_1.transformGraphQL)(parentName, hookExpr, arg, isFragment);
                if (!hookExpr.scope.hasBinding("processQuery")) {
                    const importDeclaration = core_1.types.importDeclaration([
                        core_1.types.importSpecifier(core_1.types.identifier("processQuery"), core_1.types.identifier("processQuery")),
                    ], core_1.types.stringLiteral("@futurelib/graphql-compiler"));
                    const [newPath] = programPath.unshiftContainer("body", importDeclaration);
                    newPath.get("specifiers").forEach((specifier) => {
                        programPath.scope.registerBinding("module", specifier);
                    });
                }
                const binding = hookExpr.scope.getBinding("processQuery");
                const hoistedExpression = core_1.types.arrowFunctionExpression([], core_1.types.objectExpression([
                    core_1.types.objectProperty(nameExpression, ast, true),
                    ...referenced.map((el) => core_1.types.spreadElement(core_1.types.callExpression(el, []))),
                ]));
                const assignInternalExpression = core_1.types.expressionStatement(core_1.types.assignmentExpression("=", memberInternalExpression, hoistedExpression));
                declarationRoot.insertAfter(assignInternalExpression);
                const assignExpression = core_1.types.expressionStatement(core_1.types.assignmentExpression("=", memberExpression, core_1.types.arrowFunctionExpression([], core_1.types.callExpression(binding === null || binding === void 0 ? void 0 : binding.identifier, [
                    core_1.types.callExpression(memberInternalExpression, []),
                ]))));
                declarationRoot.insertAfter(assignExpression);
            }
        }
    }
}
function createVisitor(hoistFragments, hoistQueries, nullifyLoadFunction, isDev) {
    return {
        visitor: {
            Program(programPath, state) {
                var _a, _b;
                const programHashData = [
                    (_b = (_a = state === null || state === void 0 ? void 0 : state.file) === null || _a === void 0 ? void 0 : _a.opts) === null || _b === void 0 ? void 0 : _b.filename,
                    programPath.getSource(),
                ];
                const programHash = (0, crypto_1.createHash)("md5")
                    .update(programHashData.join(","))
                    .digest("hex")
                    .slice(0, 8);
                const { loadQueryExpressions, fragmentHooks, queryHooks, } = (0, findAndMarkRelevantFunctions_1.findAndMarkRelevantFunctions)(programPath, isDev);
                const rootNodes = new Set();
                transformLoadQuery(loadQueryExpressions, nullifyLoadFunction, isDev);
                transformHooks(programPath, programHash, fragmentHooks, names_1.names.useFragment, true, hoistFragments, rootNodes, isDev);
                transformHooks(programPath, programHash, queryHooks, names_1.names.useQuery, false, hoistQueries, rootNodes, isDev);
            },
        },
    };
}
exports.createVisitor = createVisitor;
//# sourceMappingURL=visitor.js.map