import { types } from "@babel/core";
import { names } from "./names";
import { findAndMarkRelevantFunctions } from "./utils/findAndMarkRelevantFunctions";
import { markWithError } from "./utils/markUnsupported";
import { findRootOfCallExpression } from "./utils/findRootOfCallExpression";
import { findRootDeclaration } from "./utils/findRootDeclaration";
import { transformGraphQL } from "./utils/transformGraphql";
import { createHash } from "crypto";
function verifyLoadQueryCall(path, arg, isDev) {
    if (!types.isIdentifier(arg)) {
        const msg = isDev
            ? `[GraphQL Compiler Plugin] ${name} must be passed an identifier as the first argument.`
            : `Invalid ${name} call (first argument was not an identifier)`;
        markWithError(path, msg);
        return false;
    }
    return true;
}
function transformLoadQuery({ supported, unsupported }, nullify, isDev) {
    for (const loadQueryExpr of supported) {
        if (nullify) {
            loadQueryExpr.replaceWith(types.callExpression(loadQueryExpr.node.callee, [types.nullLiteral()]));
        }
        else {
            const arg = loadQueryExpr.node.arguments[0];
            if (verifyLoadQueryCall(loadQueryExpr, arg, isDev)) {
                const memberExpression = types.memberExpression(arg, types.identifier(names.literal));
                const callExpression = types.callExpression(memberExpression, []);
                loadQueryExpr.replaceWith(types.callExpression(loadQueryExpr.node.callee, [
                    callExpression,
                    ...loadQueryExpr.node.arguments.slice(1),
                ]));
            }
        }
    }
}
function verifyHook(path, arg, isDev) {
    if (!types.isTemplateLiteral(arg)) {
        const msg = isDev
            ? `[GraphQL Compiler Plugin] ${name} must be passed a template literal as the first argument.`
            : `Invalid ${name} call (first argument was not template literal)`;
        markWithError(path, msg);
        return false;
    }
    return true;
}
function transformHooks(programPath, programHash, { supported, unsupported }, name, isFragment, hoist, rootNodesFound, isDev) {
    for (const hookExpr of supported) {
        const arg = hookExpr.node.arguments[0];
        if (verifyHook(hookExpr, arg, isDev)) {
            if (!hoist) {
                hookExpr.replaceWith(types.callExpression(hookExpr.node.callee, [types.nullLiteral()]));
            }
            else {
                const { parent, parentName } = findRootDeclaration(hookExpr);
                const declarationRoot = findRootOfCallExpression(hookExpr, name, isDev, rootNodesFound);
                if (declarationRoot == null) {
                    markWithError(hookExpr, "Unable to find a root function declaration");
                    continue;
                }
                const memberExpression = types.memberExpression(types.identifier(parentName), types.identifier(names.literal));
                const memberInternalExpression = types.memberExpression(types.identifier(parentName), types.identifier(names.literInternalName));
                const callExpression = types.callExpression(memberExpression, []);
                hookExpr.replaceWith(types.callExpression(hookExpr.node.callee, [callExpression]));
                let nameExpression = types.stringLiteral("query");
                if (isFragment) {
                    const memberNameExpression = types.memberExpression(types.identifier(parentName), types.identifier(names.literName));
                    const fnHash = createHash("md5")
                        .update(parent.getSource())
                        .digest("hex")
                        .slice(0, 6);
                    const hoistedNameExpression = types.arrowFunctionExpression([], types.stringLiteral(parentName + "__" + programHash + "_" + fnHash));
                    const assignNameExpression = types.expressionStatement(types.assignmentExpression("=", memberNameExpression, hoistedNameExpression));
                    declarationRoot.insertAfter(assignNameExpression);
                    nameExpression = types.callExpression(memberNameExpression, []);
                }
                const { ast, referenced } = transformGraphQL(parentName, hookExpr, arg, isFragment);
                if (!hookExpr.scope.hasBinding("processQuery")) {
                    const importDeclaration = types.importDeclaration([
                        types.importSpecifier(types.identifier("processQuery"), types.identifier("processQuery")),
                    ], types.stringLiteral("@futurelib/graphql-compiler"));
                    const [newPath] = programPath.unshiftContainer("body", importDeclaration);
                    newPath.get("specifiers").forEach((specifier) => {
                        programPath.scope.registerBinding("module", specifier);
                    });
                }
                const binding = hookExpr.scope.getBinding("processQuery");
                const hoistedExpression = types.arrowFunctionExpression([], types.objectExpression([
                    types.objectProperty(nameExpression, ast, true),
                    ...referenced.map((el) => types.spreadElement(types.callExpression(el, []))),
                ]));
                const assignInternalExpression = types.expressionStatement(types.assignmentExpression("=", memberInternalExpression, hoistedExpression));
                declarationRoot.insertAfter(assignInternalExpression);
                const assignExpression = types.expressionStatement(types.assignmentExpression("=", memberExpression, types.arrowFunctionExpression([], types.callExpression(binding === null || binding === void 0 ? void 0 : binding.identifier, [
                    types.callExpression(memberInternalExpression, []),
                ]))));
                declarationRoot.insertAfter(assignExpression);
            }
        }
    }
}
export function createVisitor(hoistFragments, hoistQueries, nullifyLoadFunction, isDev) {
    return {
        visitor: {
            Program(programPath, state) {
                var _a, _b;
                const programHashData = [
                    (_b = (_a = state === null || state === void 0 ? void 0 : state.file) === null || _a === void 0 ? void 0 : _a.opts) === null || _b === void 0 ? void 0 : _b.filename,
                    programPath.getSource(),
                ];
                const programHash = createHash("md5")
                    .update(programHashData.join(","))
                    .digest("hex")
                    .slice(0, 8);
                const { loadQueryExpressions, fragmentHooks, queryHooks, } = findAndMarkRelevantFunctions(programPath, isDev);
                const rootNodes = new Set();
                transformLoadQuery(loadQueryExpressions, nullifyLoadFunction, isDev);
                transformHooks(programPath, programHash, fragmentHooks, names.useFragment, true, hoistFragments, rootNodes, isDev);
                transformHooks(programPath, programHash, queryHooks, names.useQuery, false, hoistQueries, rootNodes, isDev);
            },
        },
    };
}
//# sourceMappingURL=visitor.js.map